/* Zplet, a Z-Machine interpreter in Java */
/* Copyright 1996 Matthew T. Russotto */
/* All Rights Reserved */
/* You may freely copy, modify, and make derivative works from this code
   provided my copyright notice is retained and any modifications are
   clearly marked */

import java.util.*;
import java.awt.*;
import java.net.*;
import java.io.*;
import java.applet.Applet;

class ZHeader3 extends ZHeader
{
	final static int FILE_LENGTH_FACTOR = 2;

	ZHeader3 (byte [] memory_image)
	{
		this.memory_image = memory_image;
	}

	public boolean time_game() /* as opposed to score game */
	{
		return ((memory_image[FLAGS1]&0x02) == 2);
	}
	
	public void set_status_unavailable(boolean unavail) 
	{
		if (unavail) {
			memory_image[FLAGS1] |= 0x10;
		}
		else {
			memory_image[FLAGS1] &= 0xEF;
		}
	}

	public void set_splitting_available(boolean avail)
	{
		if (avail) {
			memory_image[FLAGS1] |= 0x20;
		}
		else {
			memory_image[FLAGS1] &= 0xDF;
		}
	}

	public void set_variable_default(boolean variable)
	{
		if (variable) {
			memory_image[FLAGS1] |= 0x40;
		}
		else {
			memory_image[FLAGS1] &= 0xBF;
		}
	}

	public int file_length() {
		int packed_length;
		
		packed_length = (((memory_image[FILE_LENGTH]&0xFF)<<8) |
					   (memory_image[FILE_LENGTH+1]&0xFF));
		return packed_length * FILE_LENGTH_FACTOR;
	}
}

class ZObjectTree3 extends ZObjectTree
{
	ZObjectTree3(ZMachine zm) {
		super(zm);
	}

	protected int ptableoffset()
	{
		return 7;
	}

	protected int getentryloc(short object)
	{
		return object_tree + (((object&0xFFFF) - 1) * 9);
	}
	
	protected int num_properties()
	{
		return 31;
	}

	short parent(short object) {
		int entryloc;
		
		entryloc = getentryloc(object);
		return (short)(zm.memory_image[entryloc + 4] & 0xFF);
	}

	short sibling(short object) {
		int entryloc;
		
		entryloc = getentryloc(object);
		return (short)(zm.memory_image[entryloc + 5] & 0xFF);
	}

	short child(short object) {
		int entryloc;
		
		entryloc = getentryloc(object);
		return (short)(zm.memory_image[entryloc + 6] & 0xFF);
	}

	void set_parent(short object, short newparent) {
		int entryloc;
		
		entryloc = getentryloc(object);
		zm.memory_image[entryloc + 4] = (byte)newparent;
	}

	void set_sibling(short object, short newparent) {
		int entryloc;
		
		entryloc = getentryloc(object);
		zm.memory_image[entryloc + 5] = (byte)newparent;
	}

	void set_child(short object, short newparent) {
		int entryloc;
		
		entryloc = getentryloc(object);
		zm.memory_image[entryloc + 6] = (byte)newparent;
	}
	int prop_entry_address(short object, short propnum) {
		int entry_address;
		int sizebyte;

//		System.err.println(" prop_entry_address " + object + " " +
//						   propnum);
		entry_address = property_table_addr(object);
		entry_address += (zm.memory_image[entry_address]&0xFF) * 2 + 1;
		sizebyte = zm.memory_image[entry_address]&0xFF;
		while (sizebyte != 0) {
//			System.err.println("propnum size addr " + (sizebyte&31) +
//							   " " + (sizebyte >>5) + " " + 
//							   Integer.toString(entry_address,16));
			if ((sizebyte & 31) == propnum)
				return entry_address;
			else if ((sizebyte & 31) < propnum)
				return 0;
			entry_address += (sizebyte>>5) + 2;
			sizebyte = zm.memory_image[entry_address]&0xFF;
		}
		return 0;
	}

	short next_prop(short object, short propnum) {
		int entry_address;
		int sizebyte;

		if (propnum == 0) {
			entry_address = property_table_addr(object);
			if (entry_address == 0)
				zm.fatal("Tried to get next property for object with no properties");
			entry_address += (zm.memory_image[entry_address]&0xFF)*2 + 1;
			
		}
		else {
			entry_address = prop_entry_address(object, propnum);
			if (entry_address == 0)
				zm.fatal("Tried to get next property for nonexistent property");
			entry_address += ((zm.memory_image[entry_address]&0xFF)>>5) + 2;
		}


		sizebyte = zm.memory_image[entry_address]&0xFF;
		return (short)(sizebyte & 31);
	}

	short prop_address(short object, short propnum) {
		int entry_address = prop_entry_address(object, propnum);

		if (entry_address == 0)
			return 0;
		return (short)(entry_address + 1);
	}

	short prop_len(short prop_address) {
		int sizebyte;

		if (prop_address == 0) {
			zm.fatal("Tried to find length of missing property");
			return (short)-1;
		}
		
		sizebyte = zm.memory_image[(prop_address&0xFFFF) - 1] & 0xFF;
		return (short)((sizebyte>>5)+1);
	}
}

class ZDictionary3 extends ZDictionary{
	ZMachine zm;
	String separators;
	int nentries;
	int wtable_addr;
	int dict_address;
	int entry_length;
	
	ZDictionary3(ZMachine zm, int dict_address) {
		int n, i;
		char separray[];

		this.zm = zm;
		this.dict_address = dict_address;
		n = zm.memory_image[dict_address]&0xFF;
		separray = new char[n];

		for (i = 0; i < n; i++)
			separray[i] = (char)zm.memory_image[dict_address + i + 1];

		separators = new String(separray);
//		System.err.println("separators: " + separators);
		entry_length = zm.memory_image[dict_address + n + 1];
		nentries = (zm.memory_image[dict_address + n + 2] << 8)  |
			((zm.memory_image[dict_address + n + 3]) & 0xFF);
		wtable_addr = dict_address+n+4;
	}

	ZDictionary3 (ZMachine zm) {
		this(zm, zm.header.dictionary());
	}

	boolean parse_word(int textloc, int wordloc, int wordlength, int parseloc) {
		short encword[];
		int encint;
		int dictint;
		int dictloc;
		int first = 0;
		int last = nentries - 1;
		int middle;
		int parseentry;

//		System.err.println("Entered parse_word");
		if (zm.memory_image[parseloc] == zm.memory_image[parseloc+1])
			return true;

		encword = zm.encode_word(wordloc, wordlength, 2);
		encint =
			((encword[0]&0xFFFF)<<16) |
						((encword[1])&0xFFFF);
		middle = (last + first) / 2;
		while (true) {
//			System.err.print("bsearch " + first + " " + middle + " " + last);
			dictloc = wtable_addr + (middle * entry_length);
			dictint = ((zm.memory_image[dictloc]&0xFF)<<24) |
				((zm.memory_image[dictloc + 1]&0xFF)<<16) |
					((zm.memory_image[dictloc + 2]&0xFF)<<8) |
						(zm.memory_image[dictloc + 3]&0xFF);
//			System.err.println(" " + Integer.toString(encint, 16) + " " +
//							 Integer.toString(dictint, 16));
			if (encint < dictint) {
				if (first == middle) break;
				last = middle - 1;
				middle = (first + middle)/2;
			}
			else if (encint > dictint) {
				if (last == middle) break;
				first = middle + 1;
				middle = (middle + last + 1)/2;
			}
			else
				break;
		}
		if (encint != dictint) {
			dictloc = 0;
		}
//		System.err.println("dictloc " + Integer.toString(dictloc,16));
//		System.err.println("wordlength " + Integer.toString(wordlength,16));
//		System.err.println("wordloc " + Integer.toString(wordloc - textloc,16));
		parseentry = parseloc + ((zm.memory_image[parseloc+1]&0xFF)*4) + 2;
		zm.memory_image[parseentry] = (byte)((dictloc & 0xFF00)>>8);
		zm.memory_image[parseentry + 1] = (byte)(dictloc&0xFF);
		zm.memory_image[parseentry + 2] = (byte)wordlength;
		zm.memory_image[parseentry + 3] = (byte)(wordloc - textloc + 1);
		/* +1 in V3, + 2 in V5 on line above*/
		zm.memory_image[parseloc+1]++;
		if (zm.memory_image[parseloc] == zm.memory_image[parseloc+1])
			return true;
		return false;
	}

	void tokenise(int textloc, int textlength, int parseloc) {
		int wordloc, wordlength;
		int textleft = textlength;
		int dictloc;
		char ch;
		boolean pbfull;

		if ((zm.memory_image[parseloc]&0xFF) < 1)
			zm.fatal("Parse buffer less than 1 word (6 bytes)");
		zm.memory_image[parseloc + 1] = (byte)0;
		wordloc = textloc;
		wordlength = 0;
		pbfull = false;
		while (!pbfull && (textleft-- > 0)) {
			ch = (char)zm.memory_image[wordloc + wordlength];
//			System.err.println("ch @ " + Integer.toString(wordloc + wordlength, 16) + " : " + ch);
			if (separators.indexOf(ch) != -1) {
				if (wordlength > 0)
					parse_word(textloc, wordloc, wordlength, parseloc);
				pbfull = parse_word(textloc, wordloc+wordlength, 1, parseloc);
				wordloc = wordloc + wordlength + 1;
				wordlength = 0;
			}
			else if (ch == ' ') {
				if (wordlength > 0)
					pbfull = parse_word(textloc, wordloc, wordlength, parseloc);
				wordloc = wordloc + wordlength + 1;
				wordlength = 0;
			}
			else
				wordlength++;
		}

		if (!pbfull && (wordlength > 0))
			parse_word(textloc, wordloc, wordlength, parseloc);
	}
}

class ZMachine3 extends ZMachine {

	public ZMachine3(ZScreen screen, ZStatus status_line, byte [] memory_image) {
		super(screen, status_line, memory_image);
		header = new ZHeader3(memory_image);
		objects = new ZObjectTree3(this);
		zd = new ZDictionary3(this);
		globals = header.global_table();
		window = new ZWindow[2];
		window[0] = new ZWindow(screen);
		window[1] = new ZWindow(screen);
		window[1].set_transcripting(false);
		current_window = window[0];
		zi = new ZInstruction(this);
	}

	void update_status_line() {
		ZHeader3 header = (ZHeader3)this.header;
		
		status_redirect = true;
		status_location = "";
		print_string(objects.short_name_addr(get_variable((short)16)));
		status_redirect = false;
		if (header.time_game())
			status_line.update_time_line(status_location,
										 get_variable((short)17),
										 get_variable((short)18));
		else
			status_line.update_score_line(status_location,
										  get_variable((short)17),
										  get_variable((short)18));
		status_location = null;
	}

	int string_address(short addr) {
		return (((int)addr)&0xFFFF) << 1;
	}

	int routine_address(short addr) {
		return (((int)addr)&0xFFFF) << 1;
	}

	void restart() {

		super.restart();

		window[0].moveto(0,0);
		window[1].moveto(0,0);
		window[0].resize(screen.getchars(),screen.getlines());
		window[1].resize(0,0);
		window[0].movecursor(0,window[0].getHeight()-1);
	}

	void set_header_flags() { /* at start, restart, restore */
		ZHeader3 header = (ZHeader3)this.header;

		super.set_header_flags();
		header.set_status_unavailable(false);
		header.set_splitting_available(true);
		header.set_variable_default(false);
	}
}
